/*-----------------------------------------------------------------------
	Warp-GGI

	General Graphics Interface demo
	Realtime picture 'gooing'

	main.c: main program

	written by Emmanuel Marty <emarty@mirus.fr>
	November 1st, 1997

	Released under GNU Public License
-----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <ggi/libggi.h>
#include <math.h>
#include "rawpict.h"
#include "warp.h"

/*	readpcx.c */

int	readPCX (char *name, struct raw_pict *rp);

/*	dowarp.c */

extern	struct warp *initWarp(int width, int height, int depth,
		              void *source, int srclinelen);
void	disposeWarp (struct warp *w);
void	doWarp8bpp (struct warp *w, int xw, int yw, int cw);

/*	selects the best screenmode for the picture, if possible	*/

int checkmode (int width, int height, int depth,
	       int *swidth, int *sheight, int *smode, int *spixsize) {

	if ( (width <= 320) && (height <= 200) && (depth == 8)) {
		*swidth = 320;
		*sheight = 200;
		*smode = GT_8BIT;
		*spixsize = 1;
		return 1;
	}
	
	return 0;
}

/*	error handling */

#define fail(reason) { fprintf (stderr, reason); done = -1; }

/*	main program	*/

int main (int argc, char **argv) {
	ggi_visual_t	disp = NULL;
	ggi_directbuffer_t	dbuf;
	ggi_pixellinearbuffer	*plb = NULL;
	struct raw_pict	rp;
	struct warp	*w = NULL;
	char		*picname = NULL, *src, *dest;
	int		swidth, sheight, smode, spixsize;
	int		err, step = 0, done = 0;
	int		y, width, height, stride;
	int		xw, yw, cw, tval;

	step = 0;

	while (!done) switch (step) {
		case 0:
			if (argc <= 2) {
				if (argc == 2) picname = argv[1];
				else picname = "leeloo.pcx";

				step++;
			}
			else fail ("Usage: warp [picture.pcx]\n");
			break;

		case 1:
			if ( !(err = readPCX (picname, &rp)) )
				step++;
			else switch (err) {
				case RPREAD_NOMEM:
					fail ("Not enough memory.\n");
					break;
				case RPREAD_NOFILE:
					fail ("Can't open picture.\n");
					break;
				case RPREAD_READERR:
					fail ("Error loading picture.\n");
					break;
				default:
					fail ("Unknown error reading picture.\n");
					break;
			}
			break;

		case 2:
			if (checkmode (rp.width, rp.height, rp.depth,
				&swidth, &sheight, &smode, &spixsize)) {
				if ( (w = initWarp (rp.width, rp.height,
						    rp.depth,
				                    rp.framebuf,
						    rp.width * spixsize)) )
					step ++;
				else fail ("Not enough memory.\n");
			}
			else fail ("Unsupported picture format.\n");
			break;

		case 3:
			ggiInit();

			if ( (disp = ggiOpen (NULL)) )
				step++;
			else fail ("Error opening GGI display.\n");
			break;

		case 4:
			ggiSetFocus (disp);
			if ( !(err = ggiSetGraphMode (disp, swidth, sheight,
				swidth, sheight, smode)) )
				step++;
			else {
				fprintf (stderr,
				  "Error switching to %dx%d %d bpp mode.\n",
				  swidth, sheight, rp.depth);
				done = -1;
			}
			break;

		case 5:
			dbuf = NULL;

			if ( !(err = ggiDBGetBuffer (disp, &dbuf)) )
				step++;
			else fail ("Error getting display buffer.\n");
			break;

		case 6:
			if (ggiDBGetLayout (dbuf) == blPixelLinearBuffer)
				step++;
			else fail ("Error: nonlinear display buffer.\n");
			break;

		case 7:
			if ( (plb = ggiDBGetPLB (dbuf)) )
				step++;
			else fail ("Error getting pixel linear buffer.\n");
			break;
	
		case 8:
			if (rp.depth <= 8)
				ggiSetPaletteVec (disp, 0, 1 << rp.depth,
					rp.clut);

			width = rp.width; height = rp.height;
			stride = plb->stride;

			while (ggiKbhit (disp));

			tval = 0;

			while (!ggiKbhit (disp)) {
				xw  = (int) (sin((tval+100)*M_PI/128) * 30);
				yw  = (int) (sin((tval)*M_PI/256) * -35);
				cw  = (int) (sin((tval-70)*M_PI/64) * 50);
				xw += (int) (sin((tval-10)*M_PI/512) * 40);
				yw += (int) (sin((tval+30)*M_PI/512) * 40);

				doWarp8bpp (w, xw, yw, cw);

				src = w->framebuf;
				dest = plb->write;

				/* one would wait start of vblank here. */

				for (y = 0; y < height; y++) {
					memcpy (dest, src, rp.width);
					src += width;
					dest += stride;
				}

				tval = (tval+1) & 511;
			}

			done = 1;
			break;
	}

	if (step > 3) {
		ggiClose (disp);
	}
		
	if (step > 2)
		disposeWarp (w);

	if (step > 1) {
		free (rp.framebuf);
		free (rp.clut);
	}

	if (done > 0) printf ("Warp by Emmanuel Marty <emarty@mirus.fr>\n");
	return ( (done > 0) ? 0 : 2);
}
